/*
 *  xdemorse: An application to decode Morse signals to text
 *
 *  Copyright (C) 2001  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#ifndef COMMON_H
#define COMMON_H	1

#include <gtk/gtk.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

/* Definitions of contexts (stages) in Morse decoding process */
#define MARK_SIGNAL		 0x000001 /* Count fragments of a mark element   */
#define ELEM_SPACE		 0x000002 /* Count frag. of inter-element space  */
#define CHAR_SPACE		 0x000004 /* Count fragments of inter-char space */
#define WAIT_WORD_SPACE	 0x000008 /* Wait for an inter-word space        */
#define WORD_SPACE		 0x000010 /* Count fragments of inter-word space */
#define WAIT_FOR_MARK	 0x000020 /* Count fragments of no-signal space  */
#define WAIT_FOR_SPACE	 0x000040 /* Wait for a space after a long dash  */
#define MARK_TONE		 0x000080 /* Input signal is mark tone (key down)*/

/* Definitions of flags for various actions */
#define ADAPT_SPEED		 0x000100 /* Enable speed tracking */
#define MAN_THRESHOLD	 0x000200 /* Enable manual detector threshold */
#define WORD_WRAP		 0x000400 /* Enable word wrapping  */
#define DISPLAY_SIGNAL	 0x000800 /* Display output of signal detector */
#define DISPLAY_DETECTOR 0x001000 /* Display output of signal detector */
#define REVERSE_VIDEO	 0x002000 /* Reverse waterfall video */
#define ENABLE_CAT		 0x004000 /* Enable CAT for transceiver */
#define CAT_SETUP		 0x008000 /* CAT is set up */
#define ENABLE_CAT_847	 0x010000 /* Enable CAT for FT847 */
#define ENABLE_CAT_857	 0x020000 /* Enable CAT for FT857 */
#define LONG_SPACE		 0x040000 /* Long period of space (no tone) */
#define ENABLE_RECEIVE	 0x080000 /* Enable reception of Morse code */
#define RCCONFIG_OK		 0x100000 /* xdemorserc file read ok */

/* Return values */
#define ERROR			1	 /* Error condition */
#define SUCCESS			0	 /* No error condition */

/*-------------------------------------------------------------------*/

/* Runtine configuration data */
typedef struct
{
  char
	snd_card[32], /* Sound card name */
	cap_src[32],  /* Capture source  */
	cap_vol[32];  /* Capture volume  */

  int
	channel,	/* ALSA "channel" for use in mixer setup */
	num_chn,	/* Number of audio channels (2=stereo, 1=mono) */
	use_chn,	/* Channel in use (frontleft=0, frontright=1 etc) */
	cap_lev,	/* Recording/Capture level */
	dsp_rate;	/* DSP rate (speed) samples/sec */

  /* Transceiver serial devices */
  char
	ft847_serial[32],
	ft857_serial[32];
  /* Transceiver type 1=FT847 2=FT867 0=NONE */
  int tcvr_type;

  int
	tone_freq,	/* Receiver BFO Tone freq */
	red_line,	/* Waterfall BFO Tone freq marker line */
	fft_stride; /* Stride of fft over input samples    */

  int
	unit_elem,	/* Morse unit element (dot) length  */
	max_unit,	/* Maximum length of unit element   */
	min_unit,	/* Minimum length of unit element   */
	spd_wpm,	/* Current Morse speed words/min    */
	wrd_wrp,	/* Maximum line length if word wrap enabled   */
	det_thr;	/* Signal threshold for Mark/Space separation */


  /* Signal/dsp samples buffer */
  short *buffer;
  int
	buffer_idx,	 /* Index to signal samples buffer */
	buffer_size; /* Buffer size according to stereo/mono mode */

} rc_data_t;

/*-------------------------------------------------------------------*/

/* Function prototypes produced by cproto */

/* callbacks.c */
void Error_Dialog(char *message);
/* cat.c */
gboolean Open_Tcvr_Serial(void);
gboolean Read_Tcvr_Serial(unsigned char *cmnd, int block_size);
gboolean Write_Tcvr_Command(unsigned char *cmnd);
void Close_Tcvr_Serial(void);
gboolean Tune_Tcvr(double x);
gboolean Read_Rx_Freq(int *freq);
gboolean Write_Rx_Freq(int freq);
/* decode.c */
gboolean Get_Character(char *chr);
void Adapt_Decoder(void);
char Hex_to_Ascii(int *hex_code);
/* detect.c */
gboolean Get_Fragment(void);
gboolean Initialize_Detector(void);
/* display.c */
void Display_Detector(int plot);
void Display_Signal(int plot);
void Display_Waterfall(void);
void Clear_Pixbuf(void);
/* fft.c */
void Ifft_Init(int fft_size);
void Ifft(int fft_size);
gboolean mem_alloc(void **ptr, int req);
/* main.c */
int main(int argc, char *argv[]);
gboolean Print_Character(gpointer data);
int Load_Line(char *buff, FILE *pfile, char *messg);
gboolean Load_Config(gpointer data);
void Usage(void);
int isFlagSet(int flag);
int isFlagClear(int flag);
void Set_Flag(int flag);
void Clear_Flag(int flag);
void Toggle_Flag(int flag);
void Cleanup(void);
/* sound.c */
gboolean Setup_Sound_Card(char *msg, int *err);
void Close_Capture_Handle(void);
void Close_Mixer_Handle(void);
void Close_Mixer_Handle(void);
gboolean Signal_Sample(short *sample);
gboolean Xrun_Recovery(int err);

#endif

