/*  sound.c
 *
 *  Soundcard handling functions of xdemorse application
 */

/*
 *  xdemorse: An application to decode and display
 *  Morse code signals using a computer's sound card
 *
 *  Copyright (C) 2002  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "sound.h"

/* fft in buffer */
extern int *fft_in_r;

/* Runtime config data */
extern rc_data_t rc_data;

/* ALSA pcm capture and mixer handles */
static snd_pcm_t *capture_handle = NULL;
static snd_mixer_t *mixer_handle = NULL;
static snd_pcm_hw_params_t *hw_params = NULL;

/*------------------------------------------------------------------------*/

/*  Setup_Sound_Card()
 *
 *  Sets up mixer and DSP devices
 */
  gboolean
Setup_Sound_Card( char *mesg, int *error )
{
  snd_mixer_elem_t *elem;
  snd_mixer_selem_id_t *sid;
  long cmin, cmax;


  /*** Set up pcm parameters ***/
  /* Open pcm for capture */
  *error = snd_pcm_open(
	  &capture_handle, rc_data.snd_card,
	  SND_PCM_STREAM_CAPTURE, 0 );
  if( *error < 0 )
  {
	strncat( mesg, "Cannot open audio device ", 127 );
	strncat( mesg, rc_data.snd_card, 127 );
	return( FALSE );
  }

  /* Allocate memory to hardware parameters structure */
  *error = snd_pcm_hw_params_malloc( &hw_params );
  if( *error < 0 )
  {
	strncat( mesg, "Cannot allocate hardware parameter struct", 127 );
	return( FALSE );
  }

  /* Initialize hardware parameter structure */
  *error = snd_pcm_hw_params_any( capture_handle, hw_params );
  if( *error < 0 )
  {
	strncat( mesg, "Cannot initialize hardware parameter struct", 127 );
	return( FALSE );
  }

  /* Set access type */
  *error = snd_pcm_hw_params_set_access(
	  capture_handle, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED );
  if( *error < 0 )
  {
	strncat( mesg, "Cannot set access type (RW_INTERLEAVED)", 127 );
	return( FALSE );
  }

  /* Set sample format */
  *error = snd_pcm_hw_params_set_format(
	  capture_handle, hw_params, SND_PCM_FORMAT_S16_LE );
  if( *error < 0 )
  {
	strncat( mesg, "Cannot set sample format (S16_LE)", 127 );
	return( FALSE );
  }

  /* Set sample rate */
  *error = snd_pcm_hw_params_set_rate(
	  capture_handle, hw_params, rc_data.dsp_rate, 0 );
  if( *error < 0 )
  {
	snprintf( mesg, 127,
		"Cannot set sample rate to %d", rc_data.dsp_rate );
	return( FALSE );
  }

  /* Set channel count */
  *error = snd_pcm_hw_params_set_channels(
	  capture_handle, hw_params, rc_data.num_chn );
  if( *error < 0 )
  {
	snprintf( mesg, 127,
		"Cannot set channel count to %d", rc_data.num_chn );
	return( FALSE );
  }

  /* Set number of periods */
  *error = snd_pcm_hw_params_set_periods(
	  capture_handle, hw_params, NUM_PERIODS, 0 );
  if( *error < 0)
  {
	strncat( mesg, "Cannot set number periods", 127 );
	return( FALSE );
  }

  /* Set period size */
  *error = snd_pcm_hw_params_set_period_size(
	  capture_handle, hw_params, PERIOD_SIZE, 0 );
  if( *error < 0)
  {
	strncat( mesg, "Cannot set period size", 127 );
	return( FALSE );
  }

  /* Set parameters */
  *error = snd_pcm_hw_params( capture_handle, hw_params );
  if( *error < 0 )
  {
	strncat( mesg, "Cannot set capture parameters", 127 );
	return( FALSE );
  }
  snd_pcm_hw_params_free( hw_params );
  hw_params = NULL;

  /*** Set up mixer (capture source and volume) ***/
  /* Open mixer */
  *error = snd_mixer_open( &mixer_handle, 0 );
  if( *error < 0 )
  {
	strncat( mesg, "Cannot open mixer", 127 );
	return( FALSE );
  }

  /* Attach mixer */
  *error = snd_mixer_attach( mixer_handle, rc_data.snd_card );
  if( *error < 0 )
  {
	strncat( mesg, "Cannot attach mixer to ", 127 );
	strncat( mesg, rc_data.snd_card, 127 );
	return( FALSE );
  }

  /* Register mixer */
  *error = snd_mixer_selem_register( mixer_handle, NULL, NULL );
  if( *error < 0 )
  {
	strncat( mesg, "Cannot register mixer", 127 );
	return *error;
  }

  /* Load mixer */
  *error = snd_mixer_load( mixer_handle );
  if( *error < 0 )
  {
	strncat( mesg, "Cannot load mixer", 127 );
	return( FALSE );
  }

  /* Allocate selem_id structure */
  *error = snd_mixer_selem_id_malloc( &sid );
  if( *error < 0 )
  {
	strncat( mesg, "Cannot allocate selem_id struct", 127 );
	return( FALSE );
  }

  /* Find capture selem */
  snd_mixer_selem_id_set_index( sid, 0 );
  snd_mixer_selem_id_set_name( sid, rc_data.cap_src );
  elem = snd_mixer_find_selem( mixer_handle, sid );
  if( !elem )
  {
	strncat( mesg, "Cannot find capture element ", 127 );
	strncat( mesg, rc_data.cap_src, 127 );
	*error = 0;
	snd_mixer_selem_id_free(sid);
	return( FALSE );
  }

  /* Set capture switch for cap source */
  if( snd_mixer_selem_has_capture_switch(elem) )
  {
	*error = snd_mixer_selem_set_capture_switch(
		elem, rc_data.channel, 1 );
	if( *error < 0 )
	{
	  strncat( mesg, "Cannot set capture device ", 127 );
	  strncat( mesg, rc_data.cap_src, 127 );
	  snd_mixer_selem_id_free(sid);
	  return( FALSE );
	}
  }

  /* Find capture volume selem */
  if( strcmp(rc_data.cap_vol, "--") != 0 )
  {
	snd_mixer_selem_id_set_index( sid, 0 );
	snd_mixer_selem_id_set_name( sid, rc_data.cap_vol );
	elem = snd_mixer_find_selem( mixer_handle, sid );
	if( !elem )
	{
	  strncat( mesg, "Cannot find volume element ", 127 );
	  strncat( mesg, rc_data.cap_vol, 127 );
	  Error_Dialog( mesg );
	}

	/* Set capture volume */
	if( elem && snd_mixer_selem_has_capture_volume(elem) )
	{
	  /* Change from % volume to sound card value */
	  long lev;
	  snd_mixer_selem_get_capture_volume_range( elem, &cmin, &cmax );
	  lev = cmin + ((cmax - cmin) * rc_data.cap_lev) / 100;

	  /* Set capture volume */
	  *error = snd_mixer_selem_set_capture_volume(
		  elem, rc_data.channel, lev );
	  if( *error < 0 )
	  {
		snprintf( mesg, 127,
			"Cannot set capture volume to %d\n"
			"Error: %s",
			rc_data.cap_lev, snd_strerror(*error) );
		Error_Dialog( mesg );
	  }
	}
  } /* if( strcmp(rc_data.cap_vol, "--") != 0 ) */

  snd_mixer_selem_id_free(sid);
  Close_Mixer_Handle();

  /* Prepare audio interface for use */
  *error = snd_pcm_prepare( capture_handle );
  if( *error < 0 )
  {
	strncat( mesg, "Cannot prepare audio interface", 127 );
	return( FALSE );
  }

  return( TRUE );
} /* End of Setup_Sound_Card() */

/*------------------------------------------------------------------------*/

/*
 * These two functions close the sound card handles
 */
  void
Close_Capture_Handle( void )
{
  if( capture_handle != NULL )
	snd_pcm_close( capture_handle );
  capture_handle = NULL;

  if( hw_params != NULL )
	snd_pcm_hw_params_free( hw_params );
  hw_params = NULL;
}

  void
Close_Mixer_Handle( void )
{
  if( mixer_handle != NULL )
	snd_mixer_close( mixer_handle );
  mixer_handle = NULL;
}

/*------------------------------------------------------------------------*/

/*  Get_Signal_Sample()
 *
 *  Gets the next DSP sample of the signal input
 */

  gboolean
Signal_Sample( short *sample )
{
  snd_pcm_sframes_t error;

  static int
	fft_idx = 0, /* fft input buffer idx */
	cnt = 0;	 /* Count of calls to this function */


  /* Refill dsp samples buffer when needed */
  if( rc_data.buffer_idx >= rc_data.buffer_size )
  {
	/* Start buffer index according to channel */
	rc_data.buffer_idx = rc_data.use_chn;

	/* Read audio samples from DSP, abort on error */
	error = snd_pcm_readi(
		capture_handle, rc_data.buffer,	READI_FRAMES );
	if( error != READI_FRAMES )
	{
	  char mesg[128];
	  snprintf( mesg, 127,
		  "Read from audio interface failed\n"
		  "Error: %s\n", snd_strerror(error) );
	  fprintf( stderr, mesg );

	  /* Try to recover if broken pipe or suspend */
	  if( !Xrun_Recovery(error) )
	  {
		Cleanup();
		return( FALSE );
	  }
	}
  } /* End of if( buffer_idx >= buffer_size ) */

  /* Get a sample from the buffer in int form, reduce to 8 bits.
   * The slope detector gives false results if 16-bit sampling is used */
  *sample = rc_data.buffer[rc_data.buffer_idx];

  /* Display waterfall when input buffer full */
  /* rc_data.fft_stride samples added for each input element */
  if( cnt == 0 )
	fft_in_r[fft_idx]  = *sample;
  else
	fft_in_r[fft_idx] += *sample;

  if( ++cnt >= rc_data.fft_stride )
  {
	fft_idx++;
	cnt = 0;
  }

  if( fft_idx >= FFT_SIZE )
  {
	fft_idx = 0;
	Display_Waterfall();
  }

  /* Increment according to mono/stereo mode */
  rc_data.buffer_idx += rc_data.num_chn;

  return( TRUE );
} /* Signal_Sample() */

/*------------------------------------------------------------------------*/

/* Xrun_Recovery()
 *
 * Recover from underrun (broken pipe) and suspend
 */
  gboolean
Xrun_Recovery( int error )
{
  if( error == -EPIPE )
  {
	error = snd_pcm_prepare( capture_handle );
	if( error < 0 )
	{
	  char mesg[128];
	  snprintf( mesg, 127,
		  "Cannot recover from underrun, prepare failed\n"
		  "Error: %s\n", snd_strerror(error) );
	  Error_Dialog( mesg );
	  return( FALSE );
	}
  }
  else if( error == -ESTRPIPE )
  {
	while( (error = snd_pcm_resume(capture_handle)) == -EAGAIN )
	  sleep(1);
	if( error < 0 )
	{
	  error = snd_pcm_prepare( capture_handle );
	  if( error < 0 )
	  {
		char mesg[128];
		snprintf( mesg, 127,
			"Cannot recover from suspend, prepare failed\n"
			"Error: %s\n", snd_strerror(error) );
		Error_Dialog( mesg );
		return( FALSE );
	  }
	}
  }

  return( TRUE );
} /* Xrun_Recovery() */

/*------------------------------------------------------------------------*/

